import re

from pyramid.response import Response
from pyramid.authentication import AuthTktCookieHelper
from pyramid.settings import asbool
from pyramid.renderers import render_to_response

from ..constants import Conf
from ..models import PPSsuser,PPSsgroup,PPSspermission,constants
from ..ppss_auth_utils import checkPassword

from pyramid.view import view_config,forbidden_view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound
from beaker.cache import cache_region



import os,datetime,logging
l = logging.getLogger('ppssauth')


from pyramid.security import (
    Everyone, Authenticated,
    remember,forget,
    Allow,Deny,
    Everyone,ALL_PERMISSIONS
    )


def getPrincipals(uid,request):
    groups = request.session.get('principals',[])
    l.debug("####  usergroups:{g}".format(g=groups))
    return groups

#this class build the ACL consumed by ACLRoot
#ACL is in the form: {group name: [list of permission names]} and is derived by ppss_groups and ppss_permission)
#permissions to be used in views are ppss_permission elements.
class ACLBuilder(object):
    def __init__(self,baseACL,dbsession):
        self.baseACL = baseACL
        self.session  = dbsession

    @cache_region('short_term', 'ACLpermissions')
    def buildACL(self):
        groups = self.session.query(PPSsgroup).filter(PPSsgroup.enabled==1).all()
        acl = [] 
        for group in groups:
            acl.append( (Allow,
                    str("g:"+group.name),
                    tuple([str(p.name) for p in group.permissions])  
            ) )
        acl = self.baseACL + acl
        l.warn("ACLBuilder:acl built: {acl}".format(acl = acl))
        return acl


#This class stores the acl structure used by the root factory
class ACLRoot(object):
    baseACL=[(Allow, Authenticated, 'view'),
        (Allow, 'g:'+constants.SYSADMINGROUP, ALL_PERMISSIONS)
        ]

    lastupdateACL = datetime.datetime.now()
    __acl__ = [
        (Allow, 'g:'+constants.SYSADMINPERM, ALL_PERMISSIONS)
    ]


    def __init__(self, request):
        self.request = request
        aclb = ACLBuilder(ACLRoot.baseACL,request.dbsession)
        ACLRoot.__acl__ = aclb.buildACL()


class AuthController():

    def getUserById(self,userid):
        if userid<0:
            user = PPSsuser()
        else:
            user = PPSsuser.byId(userid,self.request.dbsession)
        return user

    def __init__(self,request):
        self.request = request
        self.user = None
        activemenu = ""
        activeaction = ""
        ##request.bt = Conf['bootstrapclasses']

        try:
            mr = self.request.matched_route.name.split(":")
            if len(mr) == 3:
                activemenu = mr[1]
                activeaction = mr[2]
        except:
            pass

        #for all ops on users, get the target user
        self.userid = int(self.request.matchdict.get("elementid",
            self.request.session[Conf.sessionuser]['id'] if Conf.sessionuser in self.request.session else -1 ) )
        self.user = self.getUserById(self.userid)
        

        self.retdict = {
            'midtpl':Conf.sectiontemplateinherit,
            'supertpl':Conf.mastertemplateinherit,
            'botplinherit':Conf.botemplateinherit,
            'activemenu' : activemenu,
            'activeaction' : activeaction,
            'bc': Conf.bootstrapClasses,
            'ppsauthconf':Conf,
            'msg':""
        }

    @view_config(route_name='ppsslogin',renderer=Conf.logintemplate)
    def login(self):
        r = self.request
        postloginpage = self.request.referer if (self.request.referer and self.request.referer!=self.request.route_url("ppsslogin")) else r.route_url(Conf.postloginroute)
        l.info("postloginpage = {},referer = {}, loginurl = {}, plr = {}, is true? {}".format(
            postloginpage, 
            self.request.referer, 
            self.request.route_url("ppsslogin"),
            r.route_url(Conf.postloginroute), (self.request.referer and self.request.referer!=self.request.route_url("ppsslogin")) )   )
        self.request.session['postloginpage'] = postloginpage
        if self.user and self.user.passwordExpired():
                return HTTPFound(r.route_url("ppss:user:changepassword"))
        if r.POST:
            username = r.params.get("username",u"")
            password = r.params.get("password",u"")
            superuser = False
            res = None
            l.info("Login attempt: u={username}".format(username=username))
            if username == Conf.adminname:
                superuser = True
                if Conf.adminpass and password == Conf.adminpass:
                    u = PPSsuser.byField('username',username,r.dbsession)
                    if u:
                        res = u[0]
                    else:
                        self.retdict["msg"] = 'something went wrong with your login. Please check your informations'
                        return self.retdict
            if res is None:
                res = PPSsuser.checkLogin(username,password,r.dbsession)
            if res:

                if superuser:
                    l.info("{username} logged in as superuser".format(username=username) )
                    r.session['admin'] = True
                    r.session['principals'] = ["g:admin","g:sysadmin"]
                else:
                    l.debug("{username} logged in as normal user".format(username=username) )
                    r.session['principals'] = res.getPrincipals(res.passwordExpired()) 
                    r.session['admin'] = False
                r.dbsession.expunge(res)

                ## log the last login
                llogin = PPSsuser.byId(res.id,r.dbsession)
                llogin.lastlogin = datetime.datetime.now()  
                l.debug("last login:{}".format(llogin.lastlogin))

                r.session[Conf.sessionuser] = {'id':res.id,'name':username,'user':res}
                headers = remember(r, res.id)
                return HTTPFound(postloginpage,headers=headers)
            self.request.dbsession.query(PPSsuser).filter()
            l.warn("Login attempt failed for user {user}".format(user=username))    
            self.retdict.update({'logintpl': Conf.publictemplateinherit ,'msg':'something went wrong with your login. Please check your informations'})
            return self.retdict
        self.retdict.update({'logintpl': Conf.publictemplateinherit , 'msg':''})
        return self.retdict

    @view_config(route_name='ppsslogout')
    def logout(self):
        l.debug("logout")
        l.debug("principals = {pr}".format(pr=self.request.session.get('principals',[])  ))

        headers = forget(self.request)
        self.request.session.pop('admin',None)
        self.request.session.pop(Conf.sessionuser,None)
        self.request.session.pop('principals',None)
        
        return HTTPFound(self.request.route_url(Conf.postlogoutroute),headers=headers)

    def oauthCallBack(self):
        return Response("OK")

    def registernewuser(self):
        retdict = self.retdict
        retdict["link"] = ["",""]
        if self.request.POST:
            cancreate = True
            username = self.request.params.get("username","")
            password = self.request.params.get("password","")
            confirmnewpassword = self.request.params.get("confirmnewpassword","")
            if not username:
                retdict["msg"] +="Username can not be empty."
                cancreate = False
            if not password:
                retdict["msg"] +="Password can not be empty."
                cancreate = False
            if password != confirmnewpassword:
                retdict["msg"] +="Password check doesn't match the password."
                cancreate = False
            if not checkPassword(PPSsuser(),password):
                retdict["msg"] +="Password doesn't respect minimum constraints."
                cancreate = False
            if cancreate:
                user = PPSsuser(username = username)
                user.setPassword(password)
                for g in Conf.newusergroups:
                    group = PPSsgroup.byName(g,self.request.dbsession)
                    if group:
                        user.groups.append(group)
                self.request.dbsession.add(user)
                retdict['msg'] = "User created, please go to login page."
                retdict["link"] = ["login",self.request.route_url("ppsslogin")]

            
            HTTPFound(self.request.route_url("ppsslogin"))
        return retdict
        
    @view_config(route_name="ppss:user:changepassword",renderer=Conf.changepasswordtemplate,permission='login')
    def ppsschangepassword(self):
        l.debug("change password")
        if not Conf.sessionuser in self.request.session:
            return HTTPFound(self.request.route_url("ppsslogin"))
        message = ""
        if self.user.passwordExpired():
            message = Conf.passwordexpiredmessage
        
        retdict = {'logintpl': Conf.publictemplateinherit,'msg':message,'res':True}
        retdict.update(self.retdict)
        if self.request.POST:
            oldpassword = self.request.params.get("oldpassword")
            newpassword = self.request.params.get("newpassword")
            username = self.request.session.get(Conf.sessionuser).get("name")
            res = PPSsuser.checkLogin(username,oldpassword,self.request.dbsession)     
            if res:
                res.password = newpassword       
                
            else:
                retdict['res']=False
                retdict['msg']='password is wrong'
        return retdict
        
    def listUser(self):
        elements = self.request.dbsession.query(PPSsuser).all()
        retdict = {'elements':elements}
        retdict.update(self.retdict)
        return retdict

    def editUser(self):
        l.info("edit user")
        userid = self.userid
        user = self.user
        selfediting = False
        if self.request.matched_route.name == 'ppss:user:editself':
            userid = self.request.session[Conf.sessionuser]['id']
            user = self.getUserById(userid)
            submiturl = self.request.route_url('ppss:user:editself')
            selfediting = True
        else:
            submiturl = self.request.route_url('ppss:user:edit',elementid=userid)

        l.debug("***{id} -> {user}".format(user=user,id=userid))
        retdict = dict(self.retdict,**{'msg':"",'res':True,'userid':userid,'submiturl':submiturl} )
        if not user:
            retdict['res'] = False
            retdict['msg'] = "user not found"

        editablegroups = []
        logged_id = self.request.session[Conf.sessionuser]['id']
        logged_user = self.getUserById(logged_id)

        
        if logged_user.isSuperUser():
            editablegroups = PPSsgroup.all(self.request.dbsession)
        if logged_user.hasPermission('edituser') or selfediting:
            editablegroups =  logged_user.groups
        l.debug("editablegroups = {}".format(editablegroups))

        retdict.update({"user" : user, 'allgroups':editablegroups })

        if self.request.POST:
            username = user.username
            if userid<0:
                self.request.dbsession.add(user)
                user.username = self.request.params.get("username",user.username)
            if not user:
                return retdict
            newpassword = self.request.params.get("password","")
            if newpassword:
                
                confirmnewpassword = self.request.params.get("confirmnewpassword","")
                currentpassword = self.request.params.get("currentpassword","")

                #existing user must match older password
                if userid>=0 and (not PPSsuser.checkLogin(username,currentpassword,self.request.dbsession)): 
                    retdict['msg'] = "your current password does not match" 

                elif newpassword==confirmnewpassword:
                    l.info("*****changing password for {}".format(user.username) )
                    res = checkPassword(user,newpassword)
                    l.info("checkPassword result for user {} = {}".format(user.username, res ))
                    if res:
                        user.setPassword(newpassword)
                    else:
                        retdict['msg'] = "new password doesn't match constraints." 
                else:
                    retdict['msg'] = "new password doesn't match confirmation field." 
            user.enabled = 1 if self.request.params.get("enabled")=="1" else 0
            
            groups=map(int,self.request.params.getall("allgroups"))
            l.debug("group={groups}".format(groups=groups ))
            usergroups = [PPSsgroup.byId(groupid,self.request.dbsession) for groupid in groups if groupid in set([g.id for g in editablegroups ])]
            user.groups = usergroups
            self.request.dbsession.flush()
            #return HTTPFound(self.request.route_url('ppss:user:edit',elementid = user.id) )
            #return retdict

        return retdict

    def checkPassword(self):
        newpassword = self.request.get("newpassword","")
        if self.request.session[Conf.sessionuser]['id'] != self.userid and not self.request.session['admin']: 
            return {"res":False,"message":"Utente non autorizzato"}
            #TODO: log the incident
        if not self.user:
            return {"res":False,"message":"Utente non trovato"}
        
        res = checkPassword(self.user,newpassword)
        if res: 
            return {"res":res,"message":"ok"}
        else:
            return {"res":res,"message":Conf.passwordwrongmessage}


    def listGroup(self):
        elements = self.request.dbsession.query(PPSsgroup).all()
        return dict(self.retdict,**{'elements':elements}) 

    def editGroup(self):
        groupid = int(self.request.matchdict.get("elementid","-1"))
        retdict = dict(self.retdict,**{'msg':"",'res':True,'groupid':groupid} )
        if groupid<0:
            group = PPSsgroup()
        else:
            group = PPSsgroup.byId(groupid,self.request.dbsession)
            if not group:
                return HTTPFound(self.request.route_url('ppss:group:list'))
        retdict.update({'group':group})

        if self.request.POST:  #editing group
            if groupid<0:
                self.request.dbsession.add(group)
            group.name = self.request.params.get("name")
            group.enabled = 1 if self.request.params.get("enablecheck")=="1" else 0
            l.debug("paratri: {p}".format(p=self.request.params ) )
            l.debug("group.name={name},  group.enabled={enabled}".format(name=group.name,enabled=group.enabled))
            elements = self.request.dbsession.query(PPSsgroup).all()
            return dict(retdict,**{'elements':elements})

        elif group:
            allperm = self.request.dbsession.query(PPSspermission).all()
            users = self.request.dbsession.query(PPSsuser).all()
            return render_to_response(  Conf.editgrouptemplate,
                dict(retdict,**{'group':group,'allperm':allperm, 'users': users, 'msg':""}),
                self.request )
        #return HTTPFound(self.request.route_url("ppss:group:list") )

    def listPerm(self):
        elements = self.request.dbsession.query(PPSspermission).all()
        return dict(self.retdict,**{'elements':elements})

    def editPerm(self):
        pid = int(self.request.matchdict.get('elementid',-1) )
        if pid<0:
            perm = PPSspermission(id=pid)
        else:
            perm = PPSspermission.byId(pid,self.request.dbsession)

        if self.request.POST:
            elements = self.request.dbsession.query(PPSspermission).all()
            name = self.request.params.get("name","")
            if pid<0:
                self.request.dbsession.add(PPSspermission(name=name))
            elif perm.permtype!=1:
                perm.name = name
            else:
                res = {'res':False,'msg':"Impossibile modificare il permesso"}
                return dict(self.retdict,dict(res,**{'elements':elements}) )
            res = {'res':True,'msg':"Permesso modificato"}
            return dict(self.retdict,**dict(res,**{'elements':elements}) )
        elif perm:
            return render_to_response(  Conf.editpermtemplate,dict(self.retdict,**{'perm':perm}),self.request )
        return HTTPFound(self.request.route_url("ppss:perm:list") )

    def deletePerm(self):
        perm = PPSspermission.byId(int(self.request.matchdict.get('elementid',-1)))

        if perm and perm.permtype != 1:
            self.request.dbsession.delete(perm)
            res = {'res':True,'msg':"permesso cancellato"}
        else:
            res = {'res':False,'msg':"Impossibile cancellare il permesso"}
        elements = self.request.dbsession.query(PPSspermission).all()
        return dict(self.retdict,**dict(res,**{'elements':elements}) ) 



    def addPerm2Group(self):
        perm = PPSspermission.byId(int(self.request.matchdict.get('targetid',-1)),self.request.dbsession)
        group = PPSsgroup.byId(int(self.request.matchdict.get('elementid',-1)),self.request.dbsession)
        if not perm or not group:
            return {'res':False,"msg":"error in ids"}
        for i in group.permissions:
            if i.id == perm.id:
                return {'res':False,"msg":"already present"}
        group.permissions.append(perm)
        l.info(u"adding {perm} to {group}".format(perm=perm,group=group))
        return {'res':True,"msg":"change_perm", "groupperm":group.permdict()}

    def removePerm2Group(self):
        perm = PPSspermission.byId(int(self.request.matchdict.get('targetid',-1)),self.request.dbsession)
        group = PPSsgroup.byId(int(self.request.matchdict.get('elementid',-1)),self.request.dbsession)
        
        if perm and (perm.permtype != 'y'):  #TODO add superadmin capability to do this
            for i,p in enumerate(group.permissions):
                l.info("check {} {}".format(p.id,perm.id) )
                if p.id == perm.id:
                    l.info("match")
                    group.permissions.pop(i)
                    return {'res':True,"msg":"change_perm", "groupperm":group.permdict()}
                else:
                    l.info("no match")


        return {'res':False,'msg':"Impossibile rimuovere il permesso"}


    def addUser2Group(self):
        user  = PPSsuser.byId(int(self.request.matchdict.get('targetid',-1)),self.request.dbsession)
        group = PPSsgroup.byId(int(self.request.matchdict.get('elementid',-1)),self.request.dbsession)
        if not user or not group:
            return {'res':False,"msg":"error in ids"}
        for i in group.users:
            if i.id == user.id:
                return {'res':False,"msg":"already present"}
        group.users.append(user)
        l.info(u"adding {user} to {group}".format(user=user,group=group))
        return {'res':True,"msg":"change_user", "elements":group.userdict()}


    def removUser2Group(self):
        user  = PPSsuser.byId(int(self.request.matchdict.get('targetid',-1)),self.request.dbsession)
        group = PPSsgroup.byId(int(self.request.matchdict.get('elementid',-1)),self.request.dbsession)
        if user:  #TODO add superadmin capability to do this
            for i,p in enumerate(group.users):
                if p.id == user.id:
                    group.users.pop(i)
                    return {'res':True,"msg":"change_user", "elements":group.userdict()}
        return {'res':False,'msg':"Impossibile rimuovere il permesso"}

    def parseqstring(self,qparam):
        if qparam == "" or qparam is None:
            return ""
        qparam = " " + qparam + " "
        qparam = re.sub("[%]+", "\\%", qparam)
        qparam = re.sub("[ ]+", "%", qparam)
        return qparam




    @view_config(route_name='ppss:user:search',permission='listuser',renderer="json")
    def searchUser(self):
        qparam = self.parseqstring(self.request.params.get('q',''))
        l.debug("qparam = {qp}".format(qp=qparam))
        users = self.request.dbsession.query(PPSsuser).filter(PPSsuser.enabled==1).filter(PPSsuser.username.like(qparam)).all()
        return {'res':True,'elements':[u.todict() for u in  users]}

    @view_config(route_name='ppss:group:search',permission='listuser',renderer="json")
    def searchGroup(self):
        qparam = self.parseqstring(self.request.params.get('q',''))
        users = self.request.dbsession.query(PPSsgroup).filter(PPSsgroup.enabled==1).filter(PPSsgroup.name.like(qparam)).all()
        return {'res':True,'elements':[u.todict() for u in  users]}

    @view_config(route_name='ppss:perm:search',permission='listuser',renderer="json")
    def searchParam(self):
        qparam = self.parseqstring(self.request.params.get('q',''))
        users = self.request.dbsession.query(PPSspermission).filter(PPSspermission.name.like(qparam)).all()
        return {'res':True,'elements':[u.todict() for u in  users]}

    @view_config(route_name='test:test',permission='listuser',renderer=Conf.logintemplate)
    def testroute(self):
        return {}



from pyramid.settings import asbool, aslist
import sys
#import types
PY2 = sys.version_info[0] == 2

if PY2:
    def is_nonstr_iter(v):
        return hasattr(v, '__iter__')
else:
    def is_nonstr_iter(v):
        if isinstance(v, str):
            return False
        return hasattr(v, '__iter__')

def getBT(version = 4):
    v4 = version == 4
    return {
        'xs': 'col-' if v4 else 'col-xs-'
    }

class Conf():

    @classmethod
    def tplpath(cls, filename, layout=False):
        path = ""
        if cls.templatepackage:
            path += cls.templatepackage + ":"
        if layout:
            path += cls.templatelayoutfolder
        else:
            path += cls.templatefolder
        return path + "/" + filename + "." + cls.templateextension

    @staticmethod
    def conf2list(confval):
        if not is_nonstr_iter(confval):
            #print(repr(confval))
            confval = aslist(confval,flatten=False)
        print(repr(confval))
        return confval

    @classmethod
    def setup(cls, settings):
        cls.adminname   =settings.get("ppss_auth.adminname", "admin")
        cls.adminpass   =settings.get("ppss_auth.adminpass", "")
        cls.initdb = settings.get("ppss_auth.initdb", "true").lower() == 'true'

        # routes
        cls.loginurl = settings.get("ppss_auth.login_url", "/login")
        cls.logouturl = settings.get("ppss_auth.logout_url", "/logout")
        cls.registeruurl = settings.get("ppss_auth.register_url", "/register")

        cls.postlogoutroute = settings.get("ppss_auth.post_logout_route", "ppsslogin")
        cls.postloginroute = settings.get("ppss_auth.post_login_route", "home")
        cls.postloginfollow = settings.get("ppss_auth.post_login_follow", "true").lower() == 'true'
        
        cls.saltforhash = settings.get("ppss_auth.salt", "ImTheSaltOfThisLife")

        cls.forbiddentologin = settings.get("ppss_auth.forbidden_to_login", "true").lower() == 'true'


        #password enforcement
        cls.passwordpreviousdifferent = int(settings.get("ppss_auth.password_previuos_diff","3"))
        cls.passwordrelist = Conf.conf2list( settings.get("ppss_auth.password_relist",["[A-Z]+","[a-z]+","[0-9]+","[!,.?\\/;:_\-+*@]+"]) )
        cls.passwordwrongmessage = 'La nuova password deve essere differente dalle precedenti 3 e avere almeno una maiuscola, una minuscola, un numero e un carattere fra "!,.?\\/;:_-+*".'
        cls.passwordexpiredmessage = 'Password expired. Please change it.'
        cls.passwordexpire =  int(settings.get("ppss_auth.password_expire",0)  )


        ##template path
        cls.templatepackage =       settings.get("ppss_auth.templatepackage", "ppss_auth")
        cls.templatefolder =        settings.get("ppss_auth.templatefolder", "/templates")
        cls.templatelayoutfolder =  settings.get("ppss_auth.templatelayoutfolder", "/templates/layouts")
        cls.templateextension =     settings.get("ppss_auth.templateextension", "mako")

        #public templates
        cls.logintemplate = settings.get("ppss_auth.logintemplate", cls.tplpath("login") )
        cls.changepasswordtemplate = settings.get("ppss_auth.changepasswordtemplate", cls.tplpath("change") )
        cls.publictemplateinherit = settings.get("ppss_auth.publictemplateinherit", cls.tplpath("public", True) )
        
        #bo template
        cls.listusertemplate = settings.get("ppss_auth.listuser_template", cls.tplpath("listuser") )
        cls.editusertemplate = settings.get("ppss_auth.edituser_template", cls.tplpath("edituser") )
        cls.listgrouptemplate = settings.get("ppss_auth.listgroup_template", cls.tplpath("listgroup") )
        cls.editgrouptemplate = settings.get("ppss_auth.editgroup_template", cls.tplpath("editgroup") )
        cls.listpermtemplate = settings.get("ppss_auth.listperm_template", cls.tplpath("listperm") )
        cls.editpermtemplate = settings.get("ppss_auth.editperm_template", cls.tplpath("editperm") )
        cls.registerusertemplate = settings.get("ppss_auth.registeruser_template", cls.tplpath("registeruser") )

        cls.botemplateinherit = settings.get("ppss_auth.botemplateinherit", cls.tplpath("public", True) )
        

        #bo template inheritance
        cls.mastertemplateinherit = settings.get("ppss_auth.mastertemplateinherit", cls.tplpath("masterlayout", True) )
        cls.sectiontemplateinherit = settings.get("ppss_auth.sectiontemplateinherit", cls.tplpath("midlayout", True) )

        cls.bootstrapClasses = getBT(int(settings.get("ppss_auth.bootstrapversion",4)))

        cls.testurl = settings.get("ppss_auth.testurl", "/test")

        #functionalities
        cls.newusergroups = [x for x in map(str.strip, ( settings.get("ppss_auth.newusergroups", '' ).split(",") ) ) if x  ]


        #Setup user, groups and permissions
        cls.perm2create = Conf.conf2list( settings.get("ppss_auth.permission_list",[])  )
        cls.group2create = Conf.conf2list( settings.get("ppss_auth.group_list",[])  )
        cls.user2create = Conf.conf2list( settings.get("ppss_auth.user_list",[])  )
        cls.defaultpassword = settings.get("ppss_auth.default_password",None)  
        


        #session names
        cls.sessionuser = 'ppss_auth_loggeduser'
        cls.sessionprincipals = 'ppss_auth_principals'
        cls.sessionpermissions = 'ppss_auth_permissions'


        cls.idclient = "896890850990-kdppue1lijfhkj5s7tu8lsh7gsj99ndj.apps.googleusercontent.com"
        cls.secret = "hhuJLuYo9ZnvAVd9XopH-s8n"



from alembic import config
from alembic import script
from alembic.runtime import migration
import sqlalchemy

#import exceptions
import os
from pathlib import Path
import configparser

from ppss_auth import __path__


import logging
l = logging.getLogger('ppssauth')
    

def checkDBRevision(engine=None, session=None):
  if engine is None:
        engine = session.get_bind()
  path = Path(__file__).resolve().parents[2]
  #here = os.path.join(str(__path__[0]),"..")
  config_uri = os.path.join(path,"ppss_auth/alembic/alembic.ini")
  l.debug("loading conf from {}".format(config_uri))
  iniconfig = configparser.ConfigParser()
  iniconfig.read(config_uri)
  iniconfig["alembic"]["here"] = str(path)
  l.info("checkDBRevision: path:{}".format(path))
  #engine = sqlalchemy.create_engine(DATABASE_URL)
  alembic_cfg = config.Config(config_uri)
  script_ = script.ScriptDirectory.from_config(alembic_cfg)
  with engine.begin() as conn:
      context = migration.MigrationContext.configure(conn,opts={"version_table":"ppss_auth_alembic_version"}) #the migration on db
      dbversion = context.get_current_revision()
      migrationversion = script_.get_current_head()
      if dbversion != migrationversion:
          message = 'Current DB version:{}. HEAD migration version:{}\nUpgrade the database using\nppss_auth_upgrade_db <your pyramid ini file>.ini\n'.format(dbversion,migrationversion)
          raise Exception(message)



import re
from ppss_auth.models import Base
def exclude_ppss_auth_tables(config=None,sectionname='alembic:exclude',asadict=True,merge=True):

  def include_object(object, name, type_, reflected, compare_to):
    if type_ == "table":
        for pattern in exclude_tables:
            if pattern.match(name):
                l.info("excluded {}({}) because mathes pattern {}".format(name, type_,pattern) )
                return False
    l.info("{}({}) passed all patterns".format(name,type_))
    return True

  exclude_tables = []
  if config is not None:
    config_ = config.get_section(sectionname)
    tables_ = config_.get("pattern", None)
    l.info("building exclusions")
    if tables_ is not None:
        l.debug("tables_:{}".format(tables_))
        tables = list(
            map(lambda x: re.compile("^" + x.strip().replace("*","(.*)") + "$")
                ,str.splitlines(tables_)
            ) 
        )
        
    exclude_tables += tables
  if (config is None) or merge:
    tables = []
    for table in Base.__subclasses__():
      l.debug("name {}".format(table.__tablename__))
      tables.append (re.compile("^" + table.__tablename__ + "$") )
    exclude_tables += tables
  l.info("excluded tables:{}".format(exclude_tables))
  
  if asadict:
    return {"include_object":include_object}

  return include_object






from sqlalchemy.exc import OperationalError
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
import zope.sqlalchemy

import logging
l = logging.getLogger('ppssauth.scripts')

def get_engine(settings, prefix='sqlalchemy.'):
    url = settings['alembic'][prefix+'url']
    l.info("sql url={}".format(url ))
    return create_engine(settings['alembic'][prefix+'url'])


def get_session_factory(engine):
    factory = sessionmaker()
    factory.configure(bind=engine)
    return factory


def get_tm_session(session_factory, transaction_manager):
    dbsession = session_factory()
    zope.sqlalchemy.register(
        dbsession, transaction_manager=transaction_manager)
    return dbsession

    #return engine


import sys
PY2 = sys.version_info[0] == 2
if not PY2:
    unicode = str

from sqlalchemy import (
    Table,
    Column,
    Index,
    Integer,
    Text,
    Unicode,UnicodeText,
    DateTime,
    ForeignKey,
    desc, asc,UniqueConstraint
)
from sqlalchemy.orm import relationship, backref

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import MetaData
NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}

from datetime import datetime,timedelta

import logging,uuid

from sqlalchemy.orm import joinedload

metadata = MetaData(naming_convention=NAMING_CONVENTION)
Base = declarative_base(metadata=metadata)

from .constants import Conf
from .ppss_auth_utils.password import getPasswordDigest


l = logging.getLogger('ppssauth.models')


class constants():
    SYSADMINPERM = "sysadmin"
    SYSADMINGROUP = "sysadmin"

def __createPermissions(session,permissions):
    permmap = {}
    for p in permissions:
        perm = session.query(PPSspermission).filter(PPSspermission.name == p['name']).first()
        if not perm:
            perm = PPSspermission(name = p['name'])
            session.add(perm)
        perm.permtype = p['permtype'] if 'permtype' in p else 0
        permmap[p['name'] ] = perm
    return permmap

def __createGroups(session,groups,permissions):
    groupmap = {}
    for g in groups:
        g = g.split("=")
        gname = str.strip(g[0])
        if len(g) == 2:
            permlist = map(str.strip,g[1].split(","))
        else:
            permlist = []
        group = session.query(PPSsgroup).filter(PPSsgroup.name == gname).first()
        if not group:
            group = PPSsgroup(name = gname)
            session.add(group)
        groupmap[gname] = group
        for p in permlist:
            if not group.hasPermission(p):
                if p in permissions:
                    perm = permissions[p]
                else:
                    perm = PPSspermission(name = p,permtype = 0)
                    session.add(perm)
                    permissions[p] = perm
                group.permissions.append(perm)
    return groupmap

def __createUsers(session,users,groups,defaultpassword = None):
    usermap = {}
    for u in users:
        u = u.split("=")
        uname = str.strip(u[0])
        if len(u)==2:
            grouplist = map(str.strip,u[1].split(","))
        else:
            grouplist = []
        user = session.query(PPSsuser).filter(PPSsuser.username == uname).first()
        if not user:
            user = PPSsuser(username = uname)
            if Conf.defaultpassword:
                user.setPassword(defaultpassword if defaultpassword else Conf.defaultpassword)
            session.add(user)
        usermap[uname] = user
        for g in grouplist:
            if not user.isInGroup(g,False):
                if g in groups:
                    group = groups[g]
                else:
                    group = PPSsgroup(name = g)
                    session.add(group)
                user.groups.append(group)
    return usermap


##the caller must commit this
def initdb(session=None,createdefault=False,engine = None):
    if engine is None:
        engine = session.get_bind()
    try:
        PPSsuser.all(session)
    except:
        raise Exception("no user table. exiting!")
    if createdefault:
        l.info("creating default")
        systemperms = [#PPSspermission(name = u"admin"),
                {"name": u"edituser","permtype": 1},
                {"name": u"listuser","permtype": 1},
                {"name": u"login"   ,"permtype": 1},
                {"name": constants.SYSADMINPERM,"permtype": 1},
        ]
        permissions = [{'name':p,'permtype':1} for p in Conf.perm2create]
        for sp in systemperms:
            permissions.append(sp)

        permissionmap = __createPermissions(session,permissions)
        l.debug("permission map:{}".format(permissionmap))
        groupmap = __createGroups(session, Conf.group2create + ["useradmin=edituser,listuser,login","{}={}".format(constants.SYSADMINGROUP,constants.SYSADMINPERM)] , permissionmap)
        l.debug("group map:{}".format(groupmap))
        if Conf.adminname:
            root = __createUsers(session,[Conf.adminname+"="+constants.SYSADMINGROUP],groupmap,Conf.adminpass if Conf.adminpass else None)
        usermap = __createUsers(session, Conf.user2create, groupmap )
        l.debug("user map:{}".format(usermap))
        return 0
    
    return 0

ppssuserlkppssgroup = Table('ppssuser_lk_ppssgroup', Base.metadata,
    Column('user_id',Integer,ForeignKey('ppss_user.id', ondelete="CASCADE"), primary_key=True),
    Column('group_id',Integer,ForeignKey('ppss_group.id', ondelete="CASCADE"), primary_key=True )
)
ppssgrouplkppsspermission = Table('ppssgroup_lk_ppsspermission', Base.metadata,
    Column('group_id',Integer,ForeignKey('ppss_group.id', ondelete="CASCADE"), primary_key=True),
    Column('permission_id',Integer,ForeignKey('ppss_permission.id', ondelete="CASCADE"), primary_key=True )
)


class commonTable():
    @classmethod
    def byId(cls,id,dbsession):
        return dbsession.query(cls).filter(cls.id == id).first()

    @classmethod
    def all(cls,dbsession):
        return dbsession.query(cls).all()
    @classmethod
    def all(cls,DBSession,orderby=None,orderdir='asc'):
        q = DBSession.query(cls)
        if orderby is not None:
            if orderdir == 'asc':
                q = q.order_by(getattr(cls,orderby).asc() )
            else:
                q = q.order_by((getattr(cls,orderby).desc()  ))
        return q.all()
        #return cls.orderAll(DBSession.query(cls)).all()

    @classmethod
    def byField(cls,field,value,DBSession,orderby=None,orderdir='asc'):
        q = DBSession.query(cls).filter(getattr(cls, field)==value)
        if orderby is not None:
            if orderdir == 'asc':
                q = q.order_by(getattr(cls,orderby).asc() )
            else:
                q = q.order_by((getattr(cls,orderby).desc()  ))
        return q.all()
    
    @classmethod
    def byFields(cls,fields,DBSession,orderby=None,orderdir='asc'):
        q = DBSession.query(cls)
        for i,field in enumerate(fields):
            q = q.filter(getattr(cls, field[0])==field[1])
        if orderby is not None:
            if orderdir == 'asc':
                q = q.order_by(getattr(cls,orderby).asc() )
            else:
                q = q.order_by((getattr(cls,orderby).desc()  ))
        return q.all()


class PPSsuser(Base,commonTable):
    __tablename__   = 'ppss_user'
    id              = Column(Integer, primary_key=True)
    username        = Column(Unicode(128))
    password        = Column(Unicode(1024))
    insertdt        = Column(DateTime,default=datetime.now)
    updatedt        = Column(DateTime,default=datetime.now,onupdate=datetime.now)
    lastlogin       = Column(DateTime)
    enabled         = Column(Integer,default=1)
    magicnumber     = Column(Text(),default=uuid.uuid5(uuid.uuid4(),"bidibibodibibu" ).hex )   #Conf.saltforhash
    createdby       = Column(Integer)
    disabledby      = Column(Integer)
    passwordexpire  = Column(DateTime, default = None)

    groups = relationship("PPSsgroup",secondary=ppssuserlkppssgroup,lazy='joined',cascade="all", 
        backref=backref('users',lazy='select',order_by='PPSsuser.username'))


    superusercache = None
    permissionscache = None
    permissionsmapcache = None
    groupmapcache = None

    @classmethod
    def checkLogin(cls,user,password,dbsession):
        return PPSsuser.checkCryptedLogin(user,getPasswordDigest(password),dbsession)

    @classmethod
    def checkCryptedLogin(cls,user,password,dbsession):
        res = dbsession.query(cls).filter(cls.username==user).filter(cls.password==password).filter(cls.enabled==1) \
            .options(joinedload('groups')) \
            .options(joinedload('groups.permissions')) \
            .all()
        return res[0] if len(res)==1 else None

    def passwordExpired(self,now = None):
        if now is None:
            now = datetime.now()
        return not (self.passwordexpire is None or self.passwordexpire>now)

    def getPrincipals(self,expired=False):
        return   [str("g:"+group.name ) for group in self.groups  ] 

    def todict(self):
        return { "id": self.id, "username":self.username,"enabled":self.enabled}


    def setPassword(self,password,canexpire = True):
        dig = getPasswordDigest(password)
        self.password= dig
        if Conf.passwordexpire and canexpire:
            self.passwordexpire = datetime.now() + timedelta(days = Conf.passwordexpire)
        self.passowrdhistory.append( PPSspasswordhistory(password = dig)  )
        return self


    def getPermissions(self):
        if self.permissionscache is None:
            result = set()
            for g in self.groups:
                if g.enabled:
                    for p in g.permissions:
                        result.add( (p.id, p.name, p.permtype) )
            self.permissionscache = list(set(result))
        return self.permissionscache
        # return set([p.name for p in [g.permissions for g in self.groups if g.enabled]] )

    def getPermissionsMap(self):
        if self.permissionsmapcache is None:
            permlist = self.getPermissions()
            self.permissionsmapcache = set(permlist)
        return self.permissionsmapcache

    def isSuperUser(self):
        if self.superusercache is None:
            self.superusercache = False
            if self.username == Conf.adminname:return True
            for p in self.getPermissions():
                if p[2]==1 and (p[1] == u'superadmin' or p[1] == u'sysadmin'):
                    self.superusercache = True
                    break
        return self.superusercache

    def hasPermission(self, permission):
        if self.isSuperUser():
            return True
        return permission in self.getPermissionsMap()

    def getGroupsMap(self):
        if self.groupmapcache is None:
            grouplist = []
            for g in self.groups:
                grouplist.append(g.name)
            self.groupmapcache = set(grouplist)
        return self.groupmapcache


    def isInGroup(self,groupname,whitelie=True):
        if self.isSuperUser() and whitelie:return True
        return groupname in self.getGroupsMap()

    def __unicode__(self):
        return u"<PPSsuser ({id}-{name},{enabled})>".format(id=self.id,name=self.username, enabled=self.enabled)

    def __str__(self): 
        return self.__unicode__()

class PPSspasswordhistory(Base):
    __tablename__   = 'ppss_passwordhistory'
    id              = Column(Integer, primary_key=True)
    user_id         = Column(Integer, ForeignKey('ppss_user.id'))
    insertdt        = Column(DateTime,default=datetime.now)
    password        = Column(Unicode(1024))

    user = relationship("PPSsuser", backref=backref('passowrdhistory',order_by="desc(PPSspasswordhistory.id)" ))

class PPSsloginhistory(Base):
    __tablename__   = 'ppss_loginhistory'
    id              = Column(Integer, primary_key=True)
    user_id         = Column(Integer, ForeignKey('ppss_user.id'))
    ipaddress       = Column(Unicode(128))
    insertdt        = Column(DateTime,default=datetime.now)


class PPSsgroup(Base,commonTable):
    __tablename__   = 'ppss_group'
    id     = Column(Integer, primary_key=True)
    name   = Column(Unicode(128),unique=True)
    enabled= Column(Integer,default=1)
    permissions = relationship("PPSspermission",cascade="all",secondary=ppssgrouplkppsspermission  ,backref=backref('groups'))

    permissionsmapcache = None
    permissionscache = None

    def __unicode__(self):
        return u"<PPSsgroup {name} ({id})>".format(name=self.name,id=self.id)
    def __str__(self):
        return self.__unicode__()

    def todict(self):
        return {'id':self.id,"name":self.name}

    def userdict(self):
        return [x.todict() for x in self.users]

    def permdict(self):
        return [x.todict() for x in self.permissions]

    def getPermissionsMap(self):
        if self.permissionsmapcache is None:
            self.permissionscache = []
            for p in self.permissions:
                self.permissionscache.append(p.name)
            self.permissionsmapcache = set(self.permissionscache)
        return self.permissionsmapcache

    def hasPermission(self,permissionname):
        return permissionname in self.getPermissionsMap()

    @classmethod
    def byName(cls,name,dbsession):
        return dbsession.query(cls).filter(cls.name==name).first()

class PPSspermission(Base,commonTable):
    __tablename__   = 'ppss_permission'
    id     = Column(Integer, primary_key=True)
    name   = Column(Unicode(128))
    permtype   = Column(Integer,default=0)  #1 is for built-in permissions
    systemperm = Column(Unicode(4),default=u'y')

    def __unicode__(self):
        return u"<PPSspermission {name} ({id})>".format(name=self.name,id=self.id)
    def __str__(self):
        return self.__unicode__()

    def todict(self):
        return {'id':self.id,"name":self.name,"permtype":self.permtype}


import pkg_resources  # part of setuptools

class DBVersion(Base):
    __tablename__   = 'module_db_version'
    modulename = Column(Unicode(128),primary_key=True)
    moduleversion = Column(Unicode(64),primary_key=True)
    dbversion = Column(Unicode(64))
    insertdt  = Column(DateTime,default=datetime.now,onupdate=datetime.now)

    myname = "ppss_auth"

    @staticmethod
    def updateDB(session):
        version = session.query(DBVersion).filter(DBVersion.modulename == DBVersion.myname).first()
        moduldeversion = pkg_resources.require("ppss_auth")[0].version
        l.debug("read version in db {} module version {}".format(version,moduldeversion) )
        if version is None:
            session.add(DBVersion(modulename=DBVersion.myname,moduleversion=moduldeversion,dbversion="1.0") ) 
        else:
            version.moduleversion = moduldeversion
            version.dbversion = "1.0"


from pyramid.paster import (
    get_appsettings,
    setup_logging,
    bootstrap,
    )

import argparse
import transaction
import configparser
import sqlalchemy as sa
from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker

from ..ppss_auth_utils import scriptutils
from . alchemyutils import (
    get_engine,
    get_session_factory,
    get_tm_session
)

from ..models import PPSsuser,PPSsgroup
from ppss_auth.constants import Conf


import logging
l = logging.getLogger('ppssauth.scripts')

def buildargparser():
  parser = argparse.ArgumentParser(description='create user.')
  parser.add_argument('--username','-u', 
                    help='username',required=True)
  parser.add_argument('--password','-p',help='username',required=True)
  parser.add_argument('--groups','-g',default="",help='comma separated list of groups',required=False)
  parser.add_argument(
        'config_uri',
        help='Configuration file, e.g., development.ini',
    )
  
  return parser.parse_args()


def buildargparserupdate():
  parser = argparse.ArgumentParser(description='create user.')
  parser.add_argument('--username','-u', 
                    help='username',required=True)
  parser.add_argument('--password','-p',help='username',required=False)
  parser.add_argument('--groups','-g',default="",help='comma separated list of groups to add',required=False)
  parser.add_argument('--removegroups','-r',default="",help='comma separated list of groups to remove',required=False)
  parser.add_argument(
        'config_uri',
        help='Configuration file, e.g., development.ini',
    )
  
  return parser.parse_args()


def __create_user(dbsession,username,password,groups):
  u = dbsession.query(PPSsuser).filter(PPSsuser.username==username).first()
  if u:
    raise Exception("user '{}' already exists".format(username))
  u = PPSsuser(
    username = username,
  )
  u.setPassword(password)
  for g in groups:
    group = dbsession.query(PPSsgroup).filter(PPSsgroup.name==g).first()
    if group:
      u.groups.append(group)
  dbsession.add(u)


def __update_user(dbsession,username,password=None,groups2add=[],groups2remove=[]):
  u = dbsession.query(PPSsuser).filter(PPSsuser.username==username).first()
  if u is None:
    raise Exception("user '{}' not found".format(username))
  
  #set new password
  if password:
    u.setPassword(password)

  #remove unwanted groups
  for g in groups2remove:
    for index,mygroup in enumerate(u.groups):
      if mygroup.name == g:
        u.groups.pop(index)
        break

  #add new groups
  for g in groups2add:
    group = dbsession.query(PPSsgroup).filter(PPSsgroup.name==g).first()
    if group:
      u.groups.append(group)

  return u
  




def setup():
  config = scriptutils.configFromArgv()
  Conf.setup(config["app:main"])
  engine = scriptutils.engineFromConfig(config)
  session_factory = sessionmaker(bind=engine)
  return session_factory

def create():
  args = buildargparser()

  config = scriptutils.configFromArgv()
  Conf.setup(config["app:main"])
  engine = scriptutils.engineFromConfig(config)
  session_factory = sessionmaker(bind=engine)
  

  try:
    dbsession = session_factory()
    if dbsession:
      
      __create_user(dbsession,args.username,args.password,args.groups.split(","))
      dbsession.commit()
      print("user '{}' created".format(args.username ) )
  except Exception as e:
    print("An error occurred: {}".format(e))
    dbsession.rollback()
    l.exception("oops {}".format(e))


def update():
  args = buildargparserupdate()
  session_factory = setup()


  try:
    dbsession = session_factory()
    if dbsession:
      
      __update_user(dbsession,args.username,args.password,args.groups.split(","),args.removegroups.split(","))
      dbsession.commit()
      print("user '{}' updated".format(args.username ) )
  except Exception as e:
    print("An error occurred: {}".format(e))
    dbsession.rollback()
    l.exception("oops {}".format(e))



from sys import argv
import os
import configparser
from sqlalchemy import engine_from_config
from pyramid.paster import (
  setup_logging,
)

def configFromArgv():
  config_uri = argv[-1]
  setup_logging(config_uri)  
  config = configparser.ConfigParser() 
  with open(config_uri,"r") as inifile:
    config.read_string(inifile.read())
  
  here = os.getcwd()
  config["app:main"]["here"] = here
  return config

def engineFromConfig(config):
  engine = engine_from_config(config["app:main"], "sqlalchemy.")
  return engine




#from views.auth import ppssauthpolicy,ACLRoot,getPrincipals
from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker
import transaction
import zope.sqlalchemy

from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.authentication import SessionAuthenticationPolicy
from .constants import Conf
from .models import initdb,PPSsuser
from .routes import configRoutes
from .ppss_auth_utils.db import checkDBRevision

import logging
l = logging.getLogger('ppssauth')


from pyramid.security import (
    Everyone, Authenticated,
    remember,forget,
    Allow,
    Everyone,ALL_PERMISSIONS
    )


def initAuthDb(settings):
    engine = engine_from_config(settings, "sqlalchemy.")
    factory = sessionmaker()
    factory.configure(bind=engine)
    #dbsession = get_tm_session(session_factory, transaction.manager)
    dbsession = factory()
    zope.sqlalchemy.register(
        dbsession, transaction_manager=transaction.manager)
    try:
        checkDBRevision(engine = engine,
            session = dbsession)
    except Exception as e:
        l.exception("you need to upgrade the database.\nTo upgrade, run:\nppss_auth_upgrade_db <your pyramid ini file>.ini")
        raise e
    with transaction.manager:
        initdb(dbsession,Conf.initdb)

def getLoggedUser(request):
    uid = request.session[Conf.sessionuser]['id'] if Conf.sessionuser in request.session else False
    user = request.session[Conf.sessionuser]['user'] if Conf.sessionuser in request.session else False
    if user:
        return user
    else:
        return None
    #l.warn("*****getting logged user for {}".format(uid))
    if uid:
        user = PPSsuser.byId(uid,request.dbsession)
        return user
    else:
        pass
        #l.warn("**** session is:{}.\nI was looking for this key:{} -> {}".format(request.session,Conf.sessionuser,request.session.get(Conf.sessionuser,None)))
    return None
    


configured = False
def includeme(config):
    global configured
    if configured:
        l.debug("already configured...")
        #return
    configured = True
    #ppssauthpolicy = PPSSAuthenticationPolicy(config.get_settings())
    settings = config.get_settings()
    Conf.setup(settings)
    config.add_request_method(getLoggedUser,'loggeduser',reify=True)
    initAuthDb(settings)

    configRoutes(config,Conf)
    
    from .views.auth import getPrincipals,ACLRoot
    authz_policy = ACLAuthorizationPolicy()
    config.set_authentication_policy(SessionAuthenticationPolicy(callback=getPrincipals) )
    config.set_authorization_policy(ACLAuthorizationPolicy())
    config.set_root_factory(ACLRoot)
    config.scan("ppss_auth")
    pass



from pyramid.paster import (
    get_appsettings,
    setup_logging,
    )

from sys import argv
import os

from alembic.operations import Operations
from alembic.runtime.migration import MigrationContext
import sqlalchemy as sa

from ..ppss_auth_utils import scriptutils


import logging
l = logging.getLogger('ppss_auth')

config = None

def main():

  config = scriptutils.configFromArgv()
  engine = scriptutils.engineFromConfig(config)
  
  try:
    l.info("engine created")
    conn = engine.connect()
    ctx = MigrationContext.configure(conn)
    op = Operations(ctx)
    l.info("op created")

    op.drop_table('ppssuser_lk_ppssgroup')
    op.drop_table('ppssgroup_lk_ppsspermission')
    op.drop_table('ppss_passwordhistory')
    op.drop_table('ppss_loginhistory')
    op.drop_table('ppss_user')
    op.drop_table('ppss_permission')
    op.drop_table('ppss_group')
    op.drop_table('module_db_version')
    op.drop_table('ppss_auth_alembic_version')
    l.info("tables dropped")
    conn.commit()
    l.info("connection committed")
  except Exception as e:
    l.exception("something went wrong:{}".format(e))



import logging
l = logging.getLogger('ppssauth')

def configRoutes(config,Conf):
  from .views.auth import AuthController

  config.include("pyramid_beaker")
  config.add_static_view(  name='ppss_auth_static', path='ppss_auth_static', cache_max_age=3600)
  config.add_route('ppsslogin', Conf.loginurl)
  config.add_route('ppsslogout', Conf.logouturl)


  config.add_route('ppss:user:register', Conf.registeruurl)

  config.add_route('ppss:user:changepassword', '/password/change')
  config.add_route('ppss:user:resetpassword', '/password/change/{magicnumber}')
  
  config.add_route('ppss:user:list', '/user/list')
  config.add_route('ppss:user:editself', '/user/modify/me')
  config.add_route('ppss:user:edit', '/user/modify/{elementid}')
  config.add_route('ppss:user:search', '/user/search/')

  config.add_route('ppss:group:list', '/group/list')
  config.add_route('ppss:group:edit', '/group/modify/{elementid}')
  
  config.add_route('ppss:group:removeuser', '/group/removeuser/{elementid}/{targetid}')
  config.add_route('ppss:group:adduser', '/group/adduser/{elementid}/{targetid}')
  config.add_route('ppss:group:removeperm', '/group/removeperm/{elementid}/{targetid}')
  config.add_route('ppss:group:addperm', '/group/addperm/{elementid}/{targetid}')
  config.add_route('ppss:group:search', '/group/search/')

  config.add_route('ppss:perm:list',   '/perm/list')
  config.add_route('ppss:perm:edit',   '/perm/modify/{elementid}')
  config.add_route('ppss:perm:delete', '/perm/delete/{elementid}')
  config.add_route('ppss:perm:search', '/perm/search/')

  config.add_route('ppss:user:remove', '/user/remove/{userid}/{groupid}')
  config.add_route('ppss:user:checkpassword', '/user/checkpassword/{userid}')

  config.add_route('test:test',Conf.testurl)


  config.add_route('oauth:callback',"/oauth/callback")

  ########views

  #config.add_view(AuthController,attr='login',route_name="ppsslogin", renderer=Conf.logintemplate)
  #config.add_view(AuthController,attr='logout',route_name="ppsslogout")
  #config.add_view(AuthController,attr='ppsschangepassword',route_name="ppss:user:changepassword", 
  #    renderer=Conf.changepasswordtemplate)

  if Conf.forbiddentologin:
      config.add_forbidden_view(AuthController,attr='login',renderer=Conf.logintemplate)
  
  config.add_view(AuthController,attr='listUser',route_name="ppss:user:list",
      permission="listuser", renderer=Conf.listusertemplate)
  config.add_view(AuthController,attr='editUser',route_name="ppss:user:edit",
      permission="edituser", renderer=Conf.editusertemplate)
  config.add_view(AuthController,attr='editUser',route_name="ppss:user:editself",
      permission="login", renderer=Conf.editusertemplate)
  


  config.add_view(AuthController,attr='listGroup',route_name="ppss:group:list",
      permission="listuser", renderer=Conf.listgrouptemplate)
  config.add_view(AuthController,attr='editGroup',route_name="ppss:group:edit",
      permission="edituser", renderer=Conf.listgrouptemplate)

  config.add_view(AuthController,attr='listPerm',route_name="ppss:perm:list",
      permission="sysadmin", renderer=Conf.listpermtemplate)
  config.add_view(AuthController,attr='editPerm',route_name="ppss:perm:edit",
      permission="sysadmin", renderer=Conf.listpermtemplate)
  config.add_view(AuthController,attr='deletePerm',route_name="ppss:perm:delete",
      permission="sysadmin", renderer=Conf.listpermtemplate)

  config.add_view(AuthController,attr='checkPassword',route_name="ppss:user:checkpassword",
      permission="login", renderer=Conf.editpermtemplate)


  config.add_view(AuthController,attr='addPerm2Group',route_name='ppss:group:addperm',
    permission="sysadmin",renderer="json")
  config.add_view(AuthController,attr='removePerm2Group',route_name='ppss:group:removeperm',
    permission="sysadmin",renderer="json")
  config.add_view(AuthController,attr='addUser2Group',route_name='ppss:group:adduser',
    permission="sysadmin",renderer="json")
  config.add_view(AuthController,attr='removUser2Group',route_name='ppss:group:removeuser',
    permission="sysadmin",renderer="json")


  config.add_view(AuthController,attr='registernewuser',route_name='ppss:user:register',
    renderer=Conf.registerusertemplate)


  config.add_view(AuthController,attr='oauthCallBack',route_name='oauth:callback')






"""creation

Revision ID: f724fac25359
Revises: 
Create Date: 2020-07-01 12:11:00.121115

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f724fac25359'
down_revision = None
branch_labels = None
depends_on = None

from sqlalchemy.engine.reflection import Inspector


def _table_has_column(inspector,table, column):

    has_column = False
    for col in inspector.get_columns(table):
        if column not in col['name']:
            continue
        has_column = True
    return has_column


def upgrade():
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)
    tables = inspector.get_table_names()
    if 'module_db_version' not in tables:
        # ### commands auto generated by Alembic - please adjust! ###
        op.create_table('module_db_version',
        sa.Column('modulename', sa.Unicode(length=128), nullable=False),
        sa.Column('moduleversion', sa.Unicode(length=64), nullable=False),
        sa.Column('dbversion', sa.Unicode(length=64), nullable=True),
        sa.Column('insertdt', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('modulename', 'moduleversion', name=op.f('pk_module_db_version'))
        )
    if 'ppss_group' not in tables:
        op.create_table('ppss_group',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.Unicode(length=128), nullable=True),
        sa.Column('enabled', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_ppss_group')),
        sa.UniqueConstraint('name', name=op.f('uq_ppss_group_name'))
        )
    if 'ppss_permission' not in tables:
        op.create_table('ppss_permission',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.Unicode(length=128), nullable=True),
        sa.Column('permtype', sa.Integer(), nullable=True),
        sa.Column('systemperm', sa.Unicode(length=4), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_ppss_permission'))
        )
    if 'ppss_user' not in tables:
        op.create_table('ppss_user',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('username', sa.Unicode(length=128), nullable=True),
        sa.Column('password', sa.Unicode(length=1024), nullable=True),
        sa.Column('insertdt', sa.DateTime(), nullable=True),
        sa.Column('updatedt', sa.DateTime(), nullable=True),
        sa.Column('lastlogin', sa.DateTime(), nullable=True),
        sa.Column('enabled', sa.Integer(), nullable=True),
        sa.Column('magicnumber', sa.Text(), nullable=True),
        sa.Column('createdby', sa.Integer(), nullable=True),
        sa.Column('disabledby', sa.Integer(), nullable=True),
        sa.Column('passwordexpire', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_ppss_user'))
        )
    else:
        with op.batch_alter_table('ppss_user', schema=None) as batch_op:
            if not _table_has_column(inspector,'ppss_user','passwordexpire'):
                batch_op.add_column(sa.Column('passwordexpire', sa.Integer(), nullable=True))

    if 'ppss_loginhistory' not in tables:
        op.create_table('ppss_loginhistory',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('ipaddress', sa.Unicode(length=128), nullable=True),
        sa.Column('insertdt', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['user_id'], ['ppss_user.id'], name=op.f('fk_ppss_loginhistory_user_id_ppss_user')),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_ppss_loginhistory'))
        )
    if 'ppss_passwordhistory' not in tables:
        op.create_table('ppss_passwordhistory',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('insertdt', sa.DateTime(), nullable=True),
        sa.Column('password', sa.Unicode(length=1024), nullable=True),
        sa.ForeignKeyConstraint(['user_id'], ['ppss_user.id'], name=op.f('fk_ppss_passwordhistory_user_id_ppss_user')),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_ppss_passwordhistory'))
        )
    if 'ppssgroup_lk_ppsspermission' not in tables:
        op.create_table('ppssgroup_lk_ppsspermission',
        sa.Column('group_id', sa.Integer(), nullable=False),
        sa.Column('permission_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['group_id'], ['ppss_group.id'], name=op.f('fk_ppssgroup_lk_ppsspermission_group_id_ppss_group'), ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['permission_id'], ['ppss_permission.id'], name=op.f('fk_ppssgroup_lk_ppsspermission_permission_id_ppss_permission'), ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('group_id', 'permission_id', name=op.f('pk_ppssgroup_lk_ppsspermission'))
        )
    if 'ppssuser_lk_ppssgroup' not in tables:
        op.create_table('ppssuser_lk_ppssgroup',
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('group_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['group_id'], ['ppss_group.id'], name=op.f('fk_ppssuser_lk_ppssgroup_group_id_ppss_group'), ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['user_id'], ['ppss_user.id'], name=op.f('fk_ppssuser_lk_ppssgroup_user_id_ppss_user'), ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('user_id', 'group_id', name=op.f('pk_ppssuser_lk_ppssgroup'))
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('ppssuser_lk_ppssgroup')
    op.drop_table('ppssgroup_lk_ppsspermission')
    op.drop_table('ppss_passwordhistory')
    op.drop_table('ppss_loginhistory')
    op.drop_table('ppss_user')
    op.drop_table('ppss_permission')
    op.drop_table('ppss_group')
    op.drop_table('module_db_version')
    # ### end Alembic commands ###



import sys,os
from pyramid.paster import (
    get_appsettings,
    setup_logging,
    )
import logging
l = logging.getLogger('ppssauth.scripts')
import subprocess
from ppss_auth import __path__

def main(argv=sys.argv):
  config_uri = argv[-1]
  setup_logging(config_uri)
  inihere = os.getcwd()
  here = os.path.join(str(__path__[0]),"alembic")
  l.info("cwd:{} sqluri ini:{}".format(here,config_uri))
  subprocess.call('alembic -x sqluri={} -x cwd={} upgrade head'.format(config_uri,inihere).split(" "), cwd = here)



from .. import constants
import re
import hashlib

import os,datetime,logging
l = logging.getLogger('ppssauth')

def checkPassword(user,newpassword):
  prevpasslist = user.passowrdhistory
  passworddig = getPasswordDigest(newpassword)
  for prevpass in prevpasslist[0:constants.Conf.passwordpreviousdifferent]:
    l.debug("\n{}\n vs \n{}".format(passworddig, prevpass.password))
    if passworddig == prevpass.password:
      l.warn("password already used at {}".format(prevpass.insertdt))
      return False
  for reexp in constants.Conf.passwordrelist:
    myre = re.compile(reexp)
    if myre.search(newpassword) is None:
      return False
  return True


def getPasswordDigest(password):
  s = hashlib.sha512(password.encode('utf-8'))
  dig = s.hexdigest()
  return dig






from setuptools import setup,find_packages


import os
here = os.path.abspath(os.path.dirname(__file__))
readme = open(os.path.join(here, 'README.md'), 'r').read()
changelog = open(os.path.join(here, 'README.md'), 'r').read()

setup(name='ppss_auth',
  version='0.8.1.6',
  description='simple auth scheme for pyramid, based on Mako template and sqlalchemy backend',
  long_description=readme + "\n\n\n" + changelog,
  long_description_content_type="text/markdown",
  author='pdepmcp',
  author_email='d.cariboni@pingpongstars.it',
  license='MIT',
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Framework :: Pyramid',
    'Topic :: Internet :: WWW/HTTP :: Session',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.6',
  ],
  keywords="pyramid module authentication accelerator",
  python_requires='>=2.7',
  url='http://www.pingpongstars.it',
  install_requires=['pyramid_mako','pyramid_beaker','pyramid_jinja2','alembic' ],
  #packages=['src/test1'],
  packages=find_packages(),
  include_package_data=True,
  entry_points={
    'console_scripts': [
            'initialize_ppss_auth_db=ppss_auth.scripts.initialize_db:main',
            'ppss_auth_upgrade_db=ppss_auth.scripts.upgrade_db:main',
            'sayauth=ppss_auth.scripts.sayauth:sayitnow',
            'ppss_auth_cleanup_db=ppss_auth.scripts.cleanup_db:main',
            'ppss_auth_create_user=ppss_auth.scripts.create_user:create',
            'ppss_auth_update_user=ppss_auth.scripts.create_user:update',
    ],
  }

)





import argparse
import sys,os
import configparser
import transaction

from .. import models

from pyramid.paster import (
    get_appsettings,
    setup_logging,
    )

import logging
l = logging.getLogger('ppssauth.scripts')


from . alchemyutils import (
    get_engine,
    get_session_factory,
    get_tm_session
)


def parse_args(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'config_uri',
        help='Configuration file, e.g., development.ini',
    )
    return parser.parse_args(argv[1:])

here = "../"
def main(argv=sys.argv):
    from ppss_auth import __path__
    here = os.path.join(str(__path__[0]),"..")
    config_uri = os.path.join(here,"alembic.ini")

    setup_logging(config_uri)

    config = configparser.ConfigParser()
    config.read(config_uri)
    config["alembic"]["here"] = here
    settings = config

    engine = get_engine(settings)
    session_factory = get_session_factory(engine)
    print ("got engine and session factory")
    models.Base.metadata.create_all( engine )
    print ("created")
